simple_api_call library

A dart package for performing api calls in a much simpler way. Its able to perform GET/POST/PUT/PATCH/DELETE with Form Data and JSON support builtin. It also supports Server Sent Events.

A wrapper around dio package for simplicity

Classes

ApiResponse
Contains the Api Call response data
FormData
A class for form data request body
Options
Contains request data and other metadata fields required for API call

Enums

HttpMethod
The list of supported HTTP methods

Functions

apiCall(String url, {Options options = const Options()}) Future<ApiResponse>
Makes an API call to the specified url as per the given options.